﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices;
using System.Threading;
using System.Windows.Forms;
using Styx;
using Styx.Common;
using Styx.CommonBot;
using Styx.Helpers;
using Styx.Pathing;
using Styx.Plugins;
using Styx.TreeSharp;
using Styx.WoWInternals;
using Styx.TreeSharp;

namespace HighVoltz.HBToggle
{
    public class KeyBindClass
    {
        [DllImport("user32.dll")]
        private static extern short GetAsyncKeyState(Keys vKey);

        private Keys _keyBind;
        // these modifer bools are true if our keybind uses them. otherwise false.
        private bool _shiftkey;
        private bool _ctrlkey;
        private bool _altkey;
        public KeyBindClass() { _keyBind = new Keys(); }

        public KeyBindClass(Keys k)
        { // k holds both modifer and  key so we need to extract the modifer
            _shiftkey = (k & Keys.Shift) != 0;
            _ctrlkey = (k & Keys.Control) != 0;
            _altkey = (k & Keys.Alt) != 0;
            // now we remove the modifer flags from k and assign it to _keybind.
            _keyBind = (Keys.Shift & k) | (Keys.Control & k) | (Keys.Alt & k);
        }

        public Keys KeyBind
        {
            get { return _keyBind; }
            set
            { // check for modifer flags 
                _shiftkey = (value & Keys.Shift) != 0;
                _ctrlkey = (value & Keys.Control) != 0;
                _altkey = (value & Keys.Alt) != 0;
                // remove modifer bits
                value ^= (Keys.Shift & value) | (Keys.Control & value) | (Keys.Alt & value);
                _keyBind = value;
            }
        }

        public static bool IsKeyDown(Keys vKey) { return (GetAsyncKeyState(vKey) != 0); }

        private Keys _lastKey;

        public bool PollKeys()
        { // checks button status and returns true if our beybind is pressed.
            if (IsKeyDown(_keyBind) && _lastKey != _keyBind && (!(_shiftkey ^ IsKeyDown(Keys.ShiftKey))) &&
                (!(_ctrlkey ^ IsKeyDown(Keys.ControlKey))) && (!(_altkey ^ IsKeyDown(Keys.Menu))))
            {
                _lastKey = _keyBind;
                return true;
            }
            if (!IsKeyDown(_keyBind))
            {
                _lastKey = Keys.None;
            }
            return false;
        }
    }


    public class HBTogglePlugin : HBPlugin
    {
        private readonly FormHBToggle _configForm = new FormHBToggle();
        public override string Name { get { return "HB Toggle"; } }
        public override string Author { get { return "HighVoltz"; } }
        private readonly Version _version = new Version(1, 9);
        public override Version Version { get { return _version; } }
        public override string ButtonText { get { return "KeyBinding"; } }
        public override bool WantButton { get { return true; } }
        internal static KeyBindClass HBToggleKeybind = new KeyBindClass();
        internal static KeyBindClass MovementKeybind = new KeyBindClass();


        // making a new thread because pulse() doesn't get called when bot is paused and it doesn't get called frequently 
        // enough to get acurrate key readings. thread terminates when HB terminates.
        private Thread _keyPollThread;

        [DllImport("user32.dll")]
        private static extern IntPtr GetForegroundWindow();

        private static readonly IntPtr WowHandle = StyxWoW.Memory.Process.MainWindowHandle;
        // contains info about this plugin, used to determine if this plugin is enabled from my thread loop.
        private static PluginContainer _myPlugin;

        public override void OnButtonPress()
        {
            try
            {
                _configForm.ShowDialog();
            }
            catch (Exception ex)
            {
                Logging.WriteException(ex);
            }
        }

        readonly AutoResetEvent _threadEvent = new AutoResetEvent(false);
        public override void Pulse()
        {
            if (!_init)  // _init is set to false when plugin is disabled. Because Initialize is only called once I need to call it again after plugin has been re-enabled. 
                Initialize();

            if (!_isRunning)
                _threadEvent.WaitOne();
        }


        private bool _init;
        public override void Initialize()
        {
            try
            {
                if (!_init)
                {
                    HBToggleKeybind.KeyBind = HBToggleSettings.Instance.HBKeybind;
                    MovementKeybind.KeyBind = HBToggleSettings.Instance.MovementKeybind;
                    _keyPollThread = new Thread(ThreadLoop) { IsBackground = true };
                    Logging.Write("Starting Thread: keyPollThread");
                    _keyPollThread.Start();
                    _init = true;
                }
            }
            catch (Exception ex)
            {
                Logging.WriteException(ex);
            }
        }


        public override void Dispose()
        {
            _init = false;
            _keyPollThread.Abort();
        }

        // this is being run in a new thread
        public void ThreadLoop()
        {
            while (true)
            {
                try
                {
                    // if keybind is pressed and our wow window is the currently active window then toggle HB
                    IntPtr fgWinHwnd = GetForegroundWindow();
                    if (_myPlugin == null)
                    {
                        for (int i = 0; i < PluginManager.Plugins.Count; i++)
                        {
                            var plugin = PluginManager.Plugins[i];
                            if (plugin.Name.Equals(Name))
                            {
                                _myPlugin = plugin;
                                break;
                            }
                        }
                            //for (int i=PluginManager.Plugins.)
                            _myPlugin = PluginManager.Plugins.FirstOrDefault(p => p.Name.Equals(Name));
                    }
                    if (_myPlugin != null)
                    {
                        if (_myPlugin.Enabled)
                        {
                            if (HBToggleKeybind.PollKeys() && WowHandle == fgWinHwnd)
                                ToggleHB();
                            if (MovementKeybind.PollKeys() && WowHandle == fgWinHwnd)
                                ToggleMovement();
                        }
                        else
                        {
                            _myPlugin = null;
                            _init = false; 
                            return;
                        }
                    }
                    Thread.Sleep(50);
                }
                catch (Exception ex)
                {
                    // the 'Collection was modified; enumeration operation may not execute' happens when HB core and this thread are enumerating PluginManager.Plugins at the same time.
                    // Because there's 
                    var collectionModified = ex is System.InvalidOperationException && ex.StackTrace.Contains("System.Linq.Enumerable.FirstOrDefault");
                    if (!(ex is ThreadAbortException) && collectionModified)
                        Logging.WriteException(ex);

                }
            }
        }

        static bool _isRunning = true;

        private void ToggleHB()
        {
            if (_isRunning)
            {
                Logging.Write("Honorbuddy: Paused");
                WoWMovement.MoveStop();
                Lua.DoString("UIErrorsFrame:AddMessage(\"Paused\", 1.0, 0.4, 0.0)");
                _threadEvent.Reset();
            }
            else
            {
                Logging.Write("Honorbuddy: Running");
                Lua.DoString("UIErrorsFrame:AddMessage(\"Running\", 0.0, 1.0, 0.0)");
                _threadEvent.Set();
            }

            _isRunning = !_isRunning;
        }

        private static bool _isMovementEnabled = true;

        private static void ToggleMovement()
        {
            if (_isMovementEnabled)
            {
                Logging.Write("Movement: Off");
                Lua.DoString("UIErrorsFrame:AddMessage(\"Movement: OFF\", 1.0, 0.4, 0.0)");
                WoWMovement.MoveStop();
                Navigator.PlayerMover = new NoMover();
                Navigator.NavigationProvider.StuckHandler = new NoUnstuck();
                _isMovementEnabled = false;
            }
            else
            {
                Logging.Write("Movement: On");
                Lua.DoString("UIErrorsFrame:AddMessage(\"Movement: ON\", 0.0, 1.0, 0.0)");
                Navigator.PlayerMover = new ClickToMoveMover();
                Navigator.NavigationProvider.StuckHandler = OriginalUnstuckHandler;
                _isMovementEnabled = true;
            }
        }

        #region Embedded Type - NoMover

        public class NoMover : IPlayerMover
        {
            public void Move(WoWMovement.MovementDirection direction) { }
            public void MoveStop() { }
            public void MoveTowards(WoWPoint location)
            {
                int a = 4;
            }
        }

        #endregion

        private static readonly IStuckHandler OriginalUnstuckHandler = Navigator.NavigationProvider.StuckHandler;

        #region Embedded Type - NoUnstuck

        public class NoUnstuck : IStuckHandler
        {
            public bool IsStuck() { return false; }
            public void Reset() { }
            public void Unstick() { }
        }

        #endregion

        public class HbToggleAction : Styx.TreeSharp.Action
        {
            protected override RunStatus Run(object context)
            {
                return IsRunning ? RunStatus.Failure : RunStatus.Success;
            }
        }
    }
}

#region Settings

namespace HighVoltz
{
    public class HBToggleSettings : Settings
    {
        public static readonly HBToggleSettings Instance = new HBToggleSettings();
        private HBToggleSettings() : base(Path.Combine(Utilities.AssemblyDirectory, @"Settings\HBToggle\HB_Toggle.xml")) { }

        [Setting, Styx.Helpers.DefaultValue(0)]
        public Keys HBKeybind { get; set; }

        [Setting, Styx.Helpers.DefaultValue(0)]
        public Keys MovementKeybind { get; set; }
    }
}

#endregion